from contextlib import contextmanager
from typing import Any, Dict, Iterator

from dotcfg.collections import Config


@contextmanager
def set_temporary_config(
    temp_config: Dict[str, Any], set_location: object, set_name: str = "config"
) -> Iterator:
    """
    Temporarily sets the configuration values for the duration of the
    context manager.

    Args:
        - temp_config (Dict[str, Any]): Dictionary containing (possibly nested)
            config keys and values. Nested keys should be supplied as `.`
            delimited strings.
        - set_location (object): Object to set the new configuration on
        - set_name (str): Attribute to set the new configuration on on the
            `set_location`.

    Example:

        ```python
        # Instead of `obj`, you'd typically provide your project's
        # top level package to attach to so your config can be
        # found at `my_package.config`.

        obj = object()
        with set_temporary_config({"first": 1, "nested.two": 2}, set_location=obj):
            obj.config.first  # 1
            obj.config.nested.two  # 2
        ```
    """

    # If a `Config` object isn't found at this location, create
    # an empty one to use as the default

    old_config = getattr(set_location, set_name).copy()

    try:

        for key, value in temp_config.items():
            # Handle dot-delimited strings for keys
            cfg = getattr(set_location, set_name)
            subkeys = key.split(".")
            for subkey in subkeys[:-1]:
                cfg = cfg.setdefault(subkey, Config())
            cfg[subkeys[-1]] = value
            # setattr(set_location, set_name, cfg)

        cfg = getattr(set_location, set_name)
        setattr(set_location, set_name, cfg)
        yield cfg

    finally:
        cfg = getattr(set_location, set_name)
        cfg.clear()
        cfg.update(old_config)
        setattr(set_location, set_name, cfg)



import os
import re
from ast import literal_eval
from typing import Any, Dict, List, Optional, Union, cast

import toml

from dotcfg import collections

INTERPOLATION_REGEX = re.compile(r"\${(.[^${}]*)}")


def load_toml(path: str) -> dict:
    """
    Loads a config dictionary from TOML
    """
    return {
        key: value
        for key, value in toml.load(cast(str, interpolate_env_vars(path))).items()
    }


def interpolate_config(
    config: dict, env_var_prefix: Optional[str] = None
) -> collections.Config:
    """
    Processes the initial input configuration and replaces
    system variables ($PROJECT_PATH, etc.) as well as our
    custom variables ("${}" syntax).

    Args:
        - config (dict): Loaded data from a configuration file
        - env_var_prefix (Optional[str]): Environment variable prefix to
            load from the current environment.

    Returns:
        - collections.Config: Configuration object with values populated
            and accessible via dot notation or dictionary notation.
    """

    # Toml & other file formats support nested
    # dictionaries, so we need to flatten them out
    # to avoid recursive checking when interpolating
    flat_config = collections.dict_to_flatdict(config)

    if env_var_prefix is not None:
        env_vars = load_environment_variables(env_var_prefix)

    # Interpolate any environment variables referenced
    for key, value in list(flat_config.items()):
        value = interpolate_env_vars(value)
        if isinstance(value, str):
            value = string_to_type(value)

        flat_config[key] = value

    flat_config = replace_variable_references(flat_config)
    return cast(
        collections.Config,
        collections.flatdict_to_dict(flat_config, dct_class=collections.Config),
    )


def replace_variable_references(flat_config: dict) -> dict:
    """
    Given a dictionary (with CompoundKeys as keys), we replace
    the references to other keys (with ${} syntax) with the
    actual values

    Args:
        - flat_config (dict): A dictionary containing
            collections.CompoundKey values as keys

    Returns:
        - dict: The updated (modified) dictionary that
            replaces variable references with the values
            existing elsewhere in the configuration
    """

    def check_and_replace(flat_config: dict, value: str) -> Optional[str]:
        match = INTERPOLATION_REGEX.search(value)
        if not match:
            return None

        # the matched_string includes "${}"; the matched_key is just the inner value
        matched_string = match.group(0)
        matched_key = match.group(1)

        # get the referenced key from the config value
        ref_key = collections.CompoundKey(matched_key.split("."))
        # get the value corresponding to the referenced key
        ref_value = flat_config.get(ref_key, "")

        # The configuration's value was just a reference and nothing else
        if value == matched_string:
            return ref_value

        # The reference wasn't a valid reference, so to maintain consistency
        # with the rest of the config API, a missing reference returns an
        # empty string

        # The value could either have multiple references, or
        # be used to interpolate a larger string, such
        # as a database URI or API request header
        # Ex1) "postgresql+psycopg2://${username}:${password}@{host}:${port}/${dbname}"
        # Ex2) "Bearer ${api_token}"
        return value.replace(matched_string, str(ref_value), 1)

    output = flat_config.copy()
    keys_to_check = set(output.keys())

    # Since each string variable reference could have more than one
    # reference, we iterate 10 times to avoid recursive hells and
    # to not end up in an infinite loop.
    for _ in range(10):

        # iterate over every key and value to check if the value uses variable references
        for k in list(keys_to_check):
            value = output[k]
            # Only removing from `keys_to_check` when the value
            # can't possibly be a reference (not a string) or when
            # there's no match, because we'll eventually run out of references
            # since we're iteratively replacing 10 times.

            # if the value isn't a string, it can't be a reference, so we exit
            if not isinstance(value, (str, list)):
                keys_to_check.remove(k)
                continue

            if isinstance(value, str):
                new_value = check_and_replace(flat_config=flat_config, value=value)
                # None specifically means there was no regex match, so a regular string
                # (not a reference to a missing key)
                if new_value is None:
                    keys_to_check.remove(k)
                    continue
                output[k] = new_value
            else:
                new_values: List[Any] = []
                for item in value:
                    if not isinstance(item, str):
                        new_values.append(item)
                        continue

                    interpolated = check_and_replace(flat_config=output, value=item)
                    # Explicit None check because we need to append empty strings
                    # that denote correct variable lookup syntax but missing
                    # reference
                    if interpolated is not None:
                        new_values.append(interpolated)
                    else:
                        new_values.append(item)

                # Only remove from the interation once the
                # values are the same before and after replacing
                if value == new_values:
                    keys_to_check.remove(k)

                output[k] = new_values

    return output


def load_environment_variables(
    env_var_prefix: str,
) -> Dict[collections.CompoundKey, Any]:

    flat_config = {}

    for env_var, env_var_value in os.environ.items():
        if not env_var.startswith(env_var_prefix + "__"):
            continue

        # Strips the prefix off the environment variable
        env_var_option = env_var[len(env_var_prefix + "__") :]

        # Make sure the result has at least one delimited section
        # with a key
        if "__" not in env_var:
            continue

        # Env vars with escaped characters are interpreted as a
        # literal "\", which Python helpfuly escaped with a second "\".
        # This step makes sure that the escaped characters are properly
        # interpreted.
        value = cast(str, env_var_value.encode().decode("unicode_escape"))

        # Place the environment variable into the flat config
        # as the compound key
        config_option = collections.CompoundKey(env_var_option.lower().split("__"))
        flat_config[config_option] = string_to_type(
            cast(str, interpolate_env_vars(value))
        )

    return flat_config


def interpolate_env_vars(
    env_var: Optional[str],
) -> Optional[Union[bool, int, float, str]]:
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.

    Args:
        - env_var (Optional[str]): Value that potentially has references
            to system variables

    Returns:
        - Optional[Union[bool, int, float, str]]: Value with references
            to system variables expanded; cast to appropriate python types.
    """
    if not env_var or not isinstance(env_var, str):
        return env_var

    counter = 0

    while counter < 10:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            # if a change was made, apply string-to-type casts; otherwise leave alone
            # this is because we don't want to override TOML type-casting if this function
            # is applied to a non-interpolated value
            if counter > 1:
                interpolated = string_to_type(interpolated)  # type: ignore
            return interpolated
        else:
            env_var = interpolated
        counter += 1

    return None


def string_to_type(value: str) -> Any:
    """
    Given a value read from the environment (which is always a string),
    loads the value into the correct python primative.

    Args:
        - value (str): Value read from the environment

    Returns:
        - Any: Python primative loaded. If not loadable into a more
            fitting primative, returned as a string.
    """
    if value.upper() == "TRUE":
        return True
    elif value.upper() == "FALSE":
        return False

    try:
        return literal_eval(value)
    except Exception:
        pass

    return value


def validate_config(config: collections.Config) -> None:
    """
    Validates that the configuration file is valid.
        - keys do not shadow Config methods
    Note that this is performed when the config is first loaded, but not after.
    """

    def check_valid_keys(config: collections.Config) -> None:
        """
        Recursively check that keys do not shadow methods of the Config object
        """
        invalid_keys = dir(collections.Config)
        for k, v in config.items():
            if k in invalid_keys:
                raise ValueError(
                    (
                        f'Invalid config key: "{k}".'
                        " Using this name will overlap with the configuration object's"
                        " attribute, resulting in undefined behavior."
                    )
                )
            if isinstance(v, collections.Config):
                check_valid_keys(v)

    check_valid_keys(config)


def load_configuration(
    default_path: str, *paths: str, env_var_prefix: Optional[str] = None
) -> collections.Config:
    """
    Main entrypoint to loading a configuration set.

    Args:
        - default_path (str): Path containing location to configuration
            with default values
        - *paths (str): Positional paths that contain configuration items
            that overwrite (in priority order) the previous configuration.
        - env_var_prefix (Optional[str]): An environment variable prefix
            to read values from. Environment variables with naming convention
            "<prefix>__[<optional section>]__[<optional subsection>]__key"

    Returns:
        - collections.Config: Dictionary supporting dot access
    """
    default_config = load_toml(default_path)

    # For each specified path, we assume that the later they are in the
    # provided argument list, the higher priority they are, with
    # environment variables having the highest priority.

    for path in paths:
        try:
            config_chunk = load_toml(path)
        except FileNotFoundError as exc:
            raise FileNotFoundError(
                f"Configuration file {path} was specified but does not exist."
            ) from exc

        default_config = cast(
            dict, collections.merge_dicts(default_config, config_chunk)
        )

    config = interpolate_config(default_config, env_var_prefix=env_var_prefix)

    validate_config(config)
    return config



"""
Custom data types and functions for manipulating data types
"""

from collections.abc import MutableMapping
from typing import Any, List, Optional, Tuple, Union

from box import Box

DictLike = Union[dict, Box]


class CompoundKey(tuple):
    ...


class Config(Box):
    def copy(self) -> "Config":
        """
        Creates a recursive copy of the configuration instance.
        Each level of configuration (if there are nested configs)
        is a separate `Config` instance, and each level (and overwriting or modifying)
        does not impact other levels. However, values aren't deep copied,
        so modifications to mutable objects shared between levels
        or keys will impacts all levels / keys.

        Returns:
            - Config: New (shallow) copy of the configuration
        """

        new_config = Config()
        for key, value in self.items():
            if isinstance(value, Config):
                value = value.copy()

            new_config[key] = value

        return new_config


def merge_dicts(d1: DictLike, d2: DictLike) -> DictLike:
    """
    Updates `d1` from `d2` by replacing each `(k, v1)` pair in `d1` with the
    corresponding `(k, v2)` pair in `d2`.
    If the value of each pair is itself a dict, then the value is updated
    recursively.

    Args:
        - d1 (MutableMapping): A dictionary to be replaced
        - d2 (MutableMapping): A dictionary used for replacement

    Returns:
        - A `MutableMapping` with the two dictionary contents merged
    """

    new_dict = d1.copy()

    for k, v in d2.items():
        if isinstance(new_dict.get(k), MutableMapping) and isinstance(
            v, MutableMapping
        ):
            new_dict[k] = merge_dicts(new_dict[k], d2[k])
        else:
            new_dict[k] = d2[k]
    return new_dict


def dict_to_flatdict(dct: dict, parent: CompoundKey = None) -> dict:
    """
    Converts a (nested) dictionary to a flattened representation.
    Each key of the flat dict will be a CompoundKey tuple containing the "chain of keys"
    for the corresponding value.

    Args:
        - dct (dict): The dictionary to flatten
        - parent (CompoundKey, optional): Defaults to `None`. The parent key
        (you shouldn't need to set this)

    Returns:
        - dict: A flattened dict
    """

    items: List[Tuple[CompoundKey, Any]] = []
    parent = parent or CompoundKey()
    for k, v in dct.items():
        k_parent = CompoundKey(parent + (k,))
        if isinstance(v, dict):
            items.extend(dict_to_flatdict(v, parent=k_parent).items())
        else:
            items.append((k_parent, v))
    return dict(items)


def flatdict_to_dict(dct: dict, dct_class: Optional[type] = None) -> MutableMapping:
    """
    Converts a flattened dictionary back to a nested dictionary.

    Args:
        - dct (dict): The dictionary to be nested. Each key should be a
        `CompoundKey`, as generated by `dict_to_flatdict()`
        - dct_class (type, optional): the type of the result; defaults to `dict`

    Returns:
        - MutableMapping: A `MutableMapping` used to represent a nested dictionary
    """

    result: MutableMapping = (dct_class or dict)()
    for k, v in dct.items():
        if isinstance(k, CompoundKey):
            current_dict = result
            for ki in k[:-1]:
                current_dict = current_dict.setdefault(  # type: ignore
                    ki, (dct_class or dict)()
                )
            current_dict[k[-1]] = v
        else:
            result[k] = v

    return result
